#!/usr/bin/env python3

import time
from pathlib import Path
from typing import Iterator, NamedTuple, Sequence


def iter_python_files(project_path: Path, packages: Sequence[Path]) -> Iterator[Path]:
    if packages:
        for pkg in packages:
            yield from (p.resolve() for p in (project_path / pkg).glob("**/*.py"))
        return

    for p in project_path.glob("**/*.py"):
        yield p.resolve()


class OutputsFilepaths(NamedTuple):
    log: Path
    graph: Path


def get_outputs_filepaths(project_path: Path) -> OutputsFilepaths:
    target_dir = Path.home() / Path(".local", "py_import_cycles", "outputs")
    target_dir.mkdir(parents=True, exist_ok=True)

    filename_parts = list(project_path.parts[1:]) + [str(int(time.time()))]
    filename = Path("-".join(filename_parts).replace(".", "-"))

    return OutputsFilepaths(
        log=target_dir / filename.with_suffix(".log"),
        graph=target_dir / filename.with_suffix(".gv"),
    )



#!/usr/bin/env python3

from __future__ import annotations

import abc
from pathlib import Path
from typing import Final, Iterator, Sequence

_INIT_NAME = "__init__"


def _make_init_module_path(path: Path) -> Path:
    return path / f"{_INIT_NAME}.py"


def _make_init_module_name(name: ModuleName) -> ModuleName:
    return name.joinname(_INIT_NAME)


class ModuleName:
    """This class is inspired by pathlib.Path"""

    def __init__(self, *parts: str | ModuleName) -> None:
        self._parts: Final[tuple[str, ...]] = tuple(
            entry
            for part in parts
            for entry in (part.parts if isinstance(part, ModuleName) else part.split("."))
        )

    def __str__(self) -> str:
        return ".".join(self._parts)

    def __hash__(self) -> int:
        return hash(self._parts)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModuleName):
            return NotImplemented
        return self._parts == other._parts

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ModuleName):
            return NotImplemented
        return self._parts < other._parts

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, ModuleName):
            return NotImplemented
        return self._parts > other._parts

    def __le__(self, other: object) -> bool:
        return self < other or self == other

    def __ge__(self, other: object) -> bool:
        return self > other or self == other

    @property
    def parts(self) -> Sequence[str]:
        return self._parts

    @property
    def parent(self) -> ModuleName:
        return ModuleName(*self._parts[:-1])

    @property
    def parents(self) -> Sequence[ModuleName]:
        return [ModuleName(*self._parts[:idx]) for idx in range(1, len(self._parts))][::-1]

    def joinname(self, *names: str | ModuleName) -> ModuleName:
        return ModuleName(*self._parts, *names)


class Module(abc.ABC):
    def __init__(self, *, path: Path, name: ModuleName) -> None:
        self._validate(path, name)
        self.path = path
        self.name = name

    @abc.abstractmethod
    def _validate(self, path: Path, name: ModuleName) -> None:
        raise NotImplementedError()

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Module):
            return NotImplemented
        return self.name == other.name

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Module):
            return NotImplemented
        return self.name < other.name

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Module):
            return NotImplemented
        return self.name > other.name

    def __le__(self, other: object) -> bool:
        return self < other or self == other

    def __ge__(self, other: object) -> bool:
        return self > other or self == other


class RegularPackage(Module):
    def _validate(self, path: Path, name: ModuleName) -> None:
        if path.with_suffix("").name != _INIT_NAME:
            raise ValueError(path)

        if not name.parts or name.parts[-1] != _INIT_NAME:
            raise ValueError(name)


class NamespacePackage(Module):
    def _validate(self, path: Path, name: ModuleName) -> None:
        if path.with_suffix("").name == _INIT_NAME:
            raise ValueError(path)

        if name.parts and name.parts[-1] == _INIT_NAME:
            raise ValueError(name)


class PyModule(Module):
    def _validate(self, path: Path, name: ModuleName) -> None:
        if path.with_suffix("").name == _INIT_NAME:
            raise ValueError(path)

        if name.parts and name.parts[-1] == _INIT_NAME:
            raise ValueError(name)


class ModuleFactory:
    def __init__(self, project_path: Path, packages: list[Path]) -> None:
        self._project_path = project_path
        self._pkgs_names = self._find_package_names(project_path, packages)

    @staticmethod
    def _find_package_names(project_path: Path, packages: list[Path]) -> dict[str, Path]:
        pkgs_names: dict[str, Path] = {}
        for pkg in packages:
            if _make_init_module_path(project_path / pkg).exists():
                pkgs_names.setdefault(pkg.name, pkg)
                continue

            for parent in pkg.parents[::-1]:
                if _make_init_module_path(project_path / parent).exists():
                    pkgs_names.setdefault(parent.name, pkg)
                    break

        return pkgs_names

    def make_module_from_name(self, module_name: ModuleName) -> Module:
        def _get_sanitized_rel_module_path(module_name: ModuleName) -> Path:
            if module_name and module_name.parts[0] in self._pkgs_names:
                return Path(*self._pkgs_names[module_name.parts[0]].parts) / Path(
                    *module_name.parts[1:]
                )
            return Path(*module_name.parts)

        if module_name.parts[-1] == "*":
            # Note:
            # from a.b import *
            # - If a.b is a pkg everything from a.b.__init__.py is loaded
            # - If a.b is a mod everything from a.b.py is loaded
            # -> Getting the "right" filepath is already handled below
            module_name = module_name.parent

        module_path = self._project_path.joinpath(_get_sanitized_rel_module_path(module_name))

        if module_path.is_dir():
            if (init_module_path := _make_init_module_path(module_path)).exists():
                return RegularPackage(
                    path=init_module_path,
                    name=_make_init_module_name(module_name),
                )

            return NamespacePackage(
                path=module_path,
                name=module_name,
            )

        if (py_module_path := module_path.with_suffix(".py")).exists():
            return PyModule(
                path=py_module_path,
                name=module_name,
            )

        raise ValueError(module_name)

    def make_module_from_path(self, module_path: Path) -> Module:
        def _get_sanitized_module_name(module_path: Path) -> ModuleName:
            module_path = module_path.with_suffix("")
            for name, path in self._pkgs_names.items():
                abs_module_path = self._project_path.joinpath(path)
                if module_path.is_relative_to(abs_module_path):
                    return ModuleName(name, *module_path.relative_to(abs_module_path).parts)
            return ModuleName(*module_path.relative_to(self._project_path).parts)

        module_name = _get_sanitized_module_name(module_path)

        if module_path.is_dir():
            return NamespacePackage(
                path=module_path,
                name=module_name,
            )

        if module_path.is_file() and module_path.suffix == ".py":
            if module_path.stem == _INIT_NAME:
                return RegularPackage(
                    path=module_path,
                    name=module_name,
                )

            return PyModule(
                path=module_path,
                name=module_name,
            )

        raise ValueError(module_path)

    def make_parents_of_module(self, module: Module) -> Iterator[RegularPackage]:
        # Return the parents - ie. inits - of a module if they exist:
        # - a.b.c.__init__
        #   -> a.__init__, b.__init__
        # - a.b.c
        #   -> a.__init__, b.__init__, c.__init__
        if isinstance(module, RegularPackage):
            parents = module.name.parents[1:]
        else:
            parents = module.name.parents

        for parent in parents[::-1]:
            if isinstance(
                parent_module := self.make_module_from_name(parent),
                RegularPackage,
            ):
                yield parent_module



#!/usr/bin/env python3

import ast
import sys
from pathlib import Path
from typing import Iterator, NamedTuple, Sequence

from .log import logger
from .modules import Module, ModuleFactory, ModuleName, NamespacePackage, RegularPackage

STDLIB_OR_BUILTIN = sys.stdlib_module_names.union(sys.builtin_module_names)
ImportSTMT = ast.Import | ast.ImportFrom


class NodeVisitorImports(ast.NodeVisitor):
    def __init__(self) -> None:
        self._import_stmts: list[ImportSTMT] = []

    @property
    def import_stmts(self) -> Sequence[ImportSTMT]:
        return self._import_stmts

    def visit_Import(self, node: ast.Import) -> None:
        self._import_stmts.append(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        self._import_stmts.append(node)


class ImportStmtsParser:
    def __init__(
        self,
        module_factory: ModuleFactory,
        base_module: Module,
        import_stmts: Sequence[ImportSTMT],
    ) -> None:
        self._module_factory = module_factory
        self._base_module = base_module
        self._import_stmts = import_stmts

    def get_imports(self) -> Iterator[Module]:
        yield from self._module_factory.make_parents_of_module(self._base_module)

        for module_name in self._get_module_names():
            if not module_name.parts or module_name.parts[0] in STDLIB_OR_BUILTIN:
                continue

            try:
                module = self._module_factory.make_module_from_name(module_name)
                self._validate_module(module)
            except ValueError:
                continue

            yield module

    def _get_module_names(self) -> Iterator[ModuleName]:
        for import_stmt in self._import_stmts:
            if isinstance(import_stmt, ast.Import):
                yield from self._get_module_names_of_import_stmt(import_stmt)

            elif isinstance(import_stmt, ast.ImportFrom):
                yield from self._get_module_names_of_import_from_stmt(import_stmt)

    # -----ast.Import-----

    def _get_module_names_of_import_stmt(self, import_stmt: ast.Import) -> Iterator[ModuleName]:
        for alias in import_stmt.names:
            yield ModuleName(alias.name)

    # -----ast.ImportFrom-----

    def _get_module_names_of_import_from_stmt(
        self, import_from_stmt: ast.ImportFrom
    ) -> Iterator[ModuleName]:
        try:
            yield (anchor := self._get_anchor(import_from_stmt))
        except ValueError:
            return

        for alias in import_from_stmt.names:
            # Add packages/modules to above prefix:
            # 1 -> ../a/b/c{.py,/}
            # 2 -> ../BASE/c{.py,/}
            # 3 -> ../BASE/a/b/c{.py,/}
            # 4 -> ../BASE_PARENT/a/b/c{.py,/}
            yield anchor.joinname(alias.name)

    def _get_anchor(self, import_from_stmt: ast.ImportFrom) -> ModuleName:
        # Handle the cases:
        # 1 from a.b import c (module == "a.b", level == 0)
        #   -> Python module/package: ../a/b{.py,/}
        # 2 from . import c (module == None, level == 1)
        #   -> Python module/package: ../BASE{.py,/}
        # 3 from .a.b import c (module == "a.b", level == 1)
        #   -> Python module/package: ../BASE/a/b{.py,/}
        # 4 from ..a.b import c (module == "a.b", level == 2)
        #   -> Python module/package: ../BASE_PARENT/a/b{.py,/}
        if import_from_stmt.level == 0:
            if import_from_stmt.module:
                return ModuleName(import_from_stmt.module)
            raise ValueError(import_from_stmt.module)

        try:
            parent = self._base_module.name.parents[import_from_stmt.level - 1]
        except IndexError:
            parent = ModuleName()

        return parent.joinname(import_from_stmt.module) if import_from_stmt.module else parent

    # -----helper-----

    def _validate_module(self, module: Module) -> None:
        if isinstance(self._base_module, RegularPackage) and str(module.name).startswith(
            str(self._base_module.name.parent)
        ):
            # Importing submodules within a parent init is allowed
            raise ValueError(module)


class ImportsOfModule(NamedTuple):
    module: Module
    imports: Sequence[Module]


def visit_python_file(module_factory: ModuleFactory, path: Path) -> None | ImportsOfModule:
    try:
        module = module_factory.make_module_from_path(path)
    except ValueError:
        return None

    if isinstance(module, NamespacePackage):
        return None

    try:
        with open(module.path, encoding="utf-8") as f:
            content = f.read()
    except UnicodeDecodeError as e:
        logger.debug("Cannot read python file %s: %s", module.path, e)
        return None

    try:
        tree = ast.parse(content)
    except SyntaxError as e:
        logger.debug("Cannot visit python file %s: %s", module.path, e)
        return None

    visitor = NodeVisitorImports()
    visitor.visit(tree)

    parser = ImportStmtsParser(
        module_factory,
        module,
        visitor.import_stmts,
    )

    return ImportsOfModule(
        module,
        sorted(
            frozenset(parser.get_imports()),
            key=lambda m: tuple(m.name.parts),
            reverse=True,
        ),
    )



#!/usr/bin/env python3

from pathlib import Path

from setuptools import setup


def read(version_filepath: Path) -> str:
    with version_filepath.open("r") as fp:
        return fp.read()


def get_version(version_filepath: Path) -> str:
    for line in read(version_filepath).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise RuntimeError("Unable to find version string.")


# The directory containing this file
HERE = Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="py-import-cycles",
    version=get_version(HERE / "py_import_cycles/__init__.py"),
    description="Detect import cycles in Python projects",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/si-23/py-import-cycles",
    author="Simon Jess",
    author_email="simon86betz@yahoo.de",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.12",
    ],
    packages=["py_import_cycles"],
    include_package_data=True,
    install_requires=["graphviz", "networkx"],
    entry_points={
        "console_scripts": [
            "py_import_cycles=py_import_cycles.__main__:main",
        ]
    },
)



#!/usr/bin/env python3

from typing import Iterator, Mapping, Sequence, Set, TypeVar

from networkx import DiGraph, find_cycle
from networkx.exception import NetworkXNoCycle

from .type_defs import Comparable

T = TypeVar("T", bound=Comparable)


def depth_first_search(graph: Mapping[T, Sequence[T]]) -> Iterator[tuple[T, ...]]:
    known_cycles: Set[tuple[T, ...]] = set()

    def _make_cycle(cyclic_edges: Sequence[tuple[T, T, str]]) -> tuple[T, ...]:
        # find_cycle returns:
        # [
        #     (e0, e1, ORIENTATION),
        #     (e1, e2, ORIENTATION),
        #     ...
        #     (eX-1, eX, ORIENTATION),
        #     (eX, e0, ORIENTATION),
        # ]
        return tuple(cyclic_edges[0][:-1] + tuple(ce[1] for ce in cyclic_edges[1:-1]))

    G = DiGraph([(v, w) for v, vertices, in graph.items() for w in vertices])
    for vertex in sorted(graph):
        try:
            cyclic_edges = find_cycle(G, source=vertex, orientation="original")
        except NetworkXNoCycle:
            continue

        if (cycle := _make_cycle(cyclic_edges)) not in known_cycles:
            known_cycles.add(cycle)
            known_cycles.add(cycle[::-1])
            yield cycle



#!/usr/bin/env python3

import argparse
import logging
import sys
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Callable

from . import __version__
from .cycles import detect_cycles
from .files import get_outputs_filepaths, iter_python_files
from .graphs import make_graph
from .log import logger, setup_logging
from .modules import Module, ModuleFactory
from .visitors import visit_python_file


def _parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
        allow_abbrev=False,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="show additional information for debug purposes",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show cycles if some are found",
    )
    parser.add_argument(
        "--graph",
        action="store_true",
        help="create graphical representation",
    )
    parser.add_argument(
        "--project-path",
        required=True,
        help=(
            "path to project. If no packages are given collect all Python"
            " files from this directory."
        ),
    )
    parser.add_argument(
        "--packages",
        nargs="+",
        help="collect Python files from top-level packages",
        default=[],
    )
    parser.add_argument(
        "--strategy",
        choices=["dfs", "tarjan", "johnson"],
        default="dfs",
        help="path-based strong component algorithm",
    )
    parser.add_argument(
        "--threshold",
        type=int,
        default=0,
        help="Tolerate a certain number of cycles, ie. an upper threshold.",
    )

    return parser.parse_args()


def main() -> int:
    args = _parse_arguments()

    project_path = Path(args.project_path)
    if not project_path.exists() or not project_path.is_dir():
        sys.stderr.write(f"No such directory: {project_path}\n")
        return 1

    packages = [Path(p) for p in args.packages]

    outputs_filepaths = get_outputs_filepaths(project_path)

    setup_logging(outputs_filepaths.log, args.debug)

    logger.info("Get Python files")
    python_files = iter_python_files(project_path, packages)

    logger.info("Visit Python files, get imports by module")
    module_factory = ModuleFactory(project_path, packages)

    imports_by_module = {
        visited.module: visited.imports
        for path in python_files
        if (visited := visit_python_file(module_factory, path)) is not None
    }

    if _debug():
        logger.debug(
            "Imports by module:\n%s",
            "\n".join(_make_readable_imports_by_module(imports_by_module)),
        )

    logger.info("Detect import cycles with strategy %s", args.strategy)
    unsorted_cycles = set(detect_cycles(args.strategy, imports_by_module))

    logger.info("Sort import cycles")
    sorted_cycles = sorted(unsorted_cycles, key=lambda t: (len(t), t[0].name))

    sys.stderr.write(f"Found {len(sorted_cycles)} import cycles\n")
    _log_or_show_cycles(args.verbose, sorted_cycles)

    if args.graph:
        logger.info("Close cycle and make graph")
        make_graph(
            outputs_filepaths.graph,
            args.strategy,
            [((cycle[-1],) + cycle) for cycle in sorted_cycles],
        )

    return len(unsorted_cycles) > args.threshold


#   .--helper--------------------------------------------------------------.
#   |                    _          _                                      |
#   |                   | |__   ___| |_ __   ___ _ __                      |
#   |                   | '_ \ / _ \ | '_ \ / _ \ '__|                     |
#   |                   | | | |  __/ | |_) |  __/ |                        |
#   |                   |_| |_|\___|_| .__/ \___|_|                        |
#   |                                |_|                                   |
#   '----------------------------------------------------------------------'


def _make_readable_imports_by_module(
    imports_by_module: Mapping[Module, Sequence[Module]]
) -> Sequence[str]:
    lines = []
    for ibm, ms in imports_by_module.items():
        if ms:
            lines.append(f"  {ibm.name} imports: {', '.join(str(m.name) for m in ms)}")
    return lines


def _make_readable_cycles(
    line_handler: Callable[[int, tuple[Module, ...]], Sequence[str]],
    sorted_cycles: Sequence[tuple[Module, ...]],
) -> list[str]:
    return [
        line for nr, ic in enumerate(sorted_cycles, start=1) for line in line_handler(nr, ic) if ic
    ]


def _log_or_show_cycles(
    verbose: bool,
    sorted_cycles: Sequence[tuple[Module, ...]],
) -> None:
    if verbose:
        for line in _make_readable_cycles(
            lambda nr, ic: [f"  Cycle {nr}:"] + [f"   {m.name}" for m in ic],
            sorted_cycles,
        ):
            sys.stderr.write(f"{line}\n")

    if _debug():
        logger.debug(
            "Import cycles:\n%s",
            "\n".join(
                _make_readable_cycles(
                    lambda nr, ic: [f"  Cycle {nr}: {' > '.join(str(m.name) for m in ic)}"],
                    sorted_cycles,
                )
            ),
        )


def _debug() -> bool:
    return logger.level == logging.DEBUG
